/*****************************************************************************
MODULE: netflow.cxx
PROJECT: netflow
AUTHOR: Jeff Nielsen
DATE: April 1996
DESCRIPTION: An implementation of the Ford-Fulkerson algorithm for finding
             the maximum flow in a network.  Annotated with SAIL function
	     calls for visualizing the process.
*****************************************************************************/
#include <iostream.h>
#include <fstream.h>
#include <strstrea.h>

#include "stdjeff.h"
#include "grlist.h"        // Adjacency-list graph
#include "aqueue.ht"       // Header template for queues
#define SWAN

// NetGraph class ---------------------------------------------------------

// A NetGraph has an additional field added to each link to store the flow
struct NetEdgeLink : public EdgeLink {
   int flow;                        // Current flow on the edge
   NetEdgeLink(int vt1, int vt2, int w, int f, EdgeLink* nxt =NULL) :
      EdgeLink(vt1, vt2, w, nxt) { flow = f; }
};
typedef NetEdgeLink* NetEdge;

class NetGraph : public Graph {
private:
   void replace_edges();
public:
   NetGraph(): Graph() {}                 // Default constructor
   NetGraph(istream& is) : Graph(is)      // Stream constructor
      { if (!invalid) replace_edges(); }
   ~NetGraph();                           // Destructor
   void add_flow(int, int, int);	  // Augment flow along a path
   virtual bool add_edge(int, int, int)	  // Disable
      { cout << "Can't add edge to a NetGraph.\n"; return FALSE; }
   virtual bool remove_edge(int, int)     // Disable
	   { cout << "Can't remove edge from a NetGraph.\n"; return FALSE; }
};

//////////////////////////////////////////////////
void NetGraph::replace_edges()
//
// PURPOSE: Replace Edge objects with NetEdge objects
{
   for (int v = 0; v < numVertex; v++)
      if (list[v] != NULL) {
         Edge temp = list[v];
         Edge e = list[v] =
            new NetEdgeLink(temp->v1,temp->v2,temp->weight,0,temp->next);
         delete temp;
         for (temp = e->next; temp != NULL; temp = e->next) {
            e->next =
               new NetEdgeLink(temp->v1,temp->v2,temp->weight,0,temp->next);
            delete temp;
	    e = e->next;
         }
      }
}

//////////////////////////////////////////////////
NetGraph::~NetGraph()
//
// PURPOSE: Destructor for NetGraph class
{
   if (list != NULL)
      for (int v = 0; v < numVertex; v++)
         while (list[v] != NULL) {
            NetEdge temp = (NetEdge)list[v]; // De-allocate NetEdges
            list[v] = list[v]->next;
            delete temp;
         }
}

//////////////////////////////////////////////////
void NetGraph::add_flow(int v1, int v2, int amount)
//
// PURPOSE: Augment the flow along an edge in the NetGraph
// PARAMETERS:
//    v1: Vertex edge comes from
//    v2: Vertex edge goes to
//    amount: amount to increase the flow
{
   Edge e, opp;

   if ((e = getEdge(v1, v2)) != NULL) {         // Make sure edge exists
      if ((opp = getEdge(v2, v1)) != NULL && NetEdge(opp)->flow > 0) {
         // There is an edge in other direction with positive flow
         amount -= ((NetEdge)opp)->flow;  // Take flow from opposite
         if (amount < 0) {                // direction first
            ((NetEdge)opp)->flow += amount;
            amount = 0;
         }
         else ((NetEdge)opp)->flow = 0;            
      }
      NetEdge(e)->flow += amount;
      assert(NetEdge(e)->flow <= e->weight);  // Make sure below capacity
   }
}

// Swan functions -------------------------------------------------------------

#ifdef SWAN
#include "sail.h"

// Names for two Swan graphis in this visualization
GRAPHID orig_swgraph = GRAPHID(100);
GRAPHID resid_swgraph = GRAPHID(101);

char memstr[150];
ostrstream memstm(memstr, 150);	    // Global in-memory stream for formatting

//////////////////////////////////////////////////
inline void make_swangraph(GRAPHID gid, NODETYPE nt, bool dir_undir,
			   LAYOUTCOMPTYPE lctype)
// PURPOSE: Macro for making SWAN graphs
// PARAMETERS:
//    gid: The graphid you want for this graph
//    nt: The type of node you want the graph to have
//    dir_undir: DIRECTED or UNDIRECTED graph
//    lctype: The default layout component you want this graph to have.
{
   gid = sw_newgraph(gid, nt, dir_undir);
   LCID lc = sw_newlc(gid, NULLLCID, lctype);
   sw_setcurlc(gid, lc);
}

//////////////////////////////////////////////////
void creategraph(Graph& g, GRAPHID gid)
// PURPOSE: Draws one of the two graphs in the Swan window--specialized to be
//	    in a particular shape.
// PARAMETER;
//    g: The graph to draw--this function expects no more than 8 vertices
//    gid: Graph ID of the graph in the SWAN window
{
   make_swangraph(gid, CIRCLE, DIRECTED, MANUAL);
   sw_setgraphattr(gid, GLAYOUTMODE, MANULAYOUT);
   if (gid == orig_swgraph) sw_setgraphpos(gid, 30, 10);
   else sw_setgraphpos(gid, 30, 220);

   int v;
   // Insert nodes
   for (v = 0; v < g.n(); v++) {
      memstm.seekp(0);
      memstm << v << ends;
      sw_insertnode(gid, NODEID(v), memstr);
      switch(v) {				// Place nodes myself
	 case 0:  sw_setnodeattr(gid, NODEID(v), GNODEPOS, 0, 75); break;
	 case 1:  sw_setnodeattr(gid, NODEID(v), GNODEPOS, 100, 0); break;
 	 case 2:  sw_setnodeattr(gid, NODEID(v), GNODEPOS, 100, 150); break;
	 case 3:  sw_setnodeattr(gid, NODEID(v), GNODEPOS, 200, 30); break;
	 case 4:  sw_setnodeattr(gid, NODEID(v), GNODEPOS, 200, 120); break;
	 case 5:  sw_setnodeattr(gid, NODEID(v), GNODEPOS, 300, 0); break;
	 case 6:  sw_setnodeattr(gid, NODEID(v), GNODEPOS, 300, 150); break;
	 case 7:  sw_setnodeattr(gid, NODEID(v), GNODEPOS, 400, 75); break;
         default:
	    cout << "Error, can't have more than 8 vertices for this demo.\n";
	    exit(-1);
      }
   }
   for (v = 0; v < g.n(); v++)
      for (Edge e = g.first(v); g.isEdge(e); e = g.next(e) ) {
	 memstm.seekp(0);
	 if (gid == orig_swgraph && NetEdge(e)->flow > 0)
	    memstm << NetEdge(e)->flow << "/";
	 memstm << e->weight << ends;
	 sw_insertedge(gid, NODEID(e->v1), NODEID(e->v2), memstr);
      }
}

#endif

// Network flow functions ----------------------------------------------------

//////////////////////////////////////////////////
bool find_path(Graph &g, int goal)
//
// PURPOSE: Find a path in the graph from vertex 0--based on BFS
// PARAMETERS:
//    g: Graph in which to find a path
//    goal: Vertex you're trying to reach from vertex 0
{
   Queue<int> q;			// Queue to hold vertices

   for (int v=1; v < g.n(); v++) {
      g.mark[v] = UNVISITED;
      g.pred[v] = UNDEF;
   }
   g.mark[0] = VISITED;       // Start search from vertex 0 (color gray)
   g.pred[0] = UNDEF;

   q.enqueue(0);	      // Put first vertex on queue
   while (!q.isEmpty()) {
      int u = q.dequeue();    // Next vertex to explore
      for (Edge e = g.first(u); g.isEdge(e); e = g.next(e)) {
         int w = e->v2;       // Next neighbor of vertex u
         if (g.mark[w] == UNVISITED) {
            g.mark[w] = VISITED;		// Color gray
            g.pred[w] = u;			// Mark predecessor
            if (w == goal) return TRUE;		// Found a path
            else q.enqueue(w);
         }
      }
      // Color vertex u black here--we're done with it
   }
   return FALSE;  // No path found to goal
}

//////////////////////////////////////////////////
void ford_fulkerson(NetGraph &orig, int sink)
//
// PURPOSE: Ford/Fulkerson algorithm from Cormen, et al, p. 594
// PARAMETERS:
//    orig: The original network graph
//    sink: vertex to use as the sink
{
   Graph residual = orig;   // Create a residual graph from original graph
#ifdef SWAN
   creategraph(orig, orig_swgraph);
   creategraph(residual, resid_swgraph);
   sw_displayallgraphs();
   sw_print("Here is the original network (top) and the residual "
	    "network (bottom).");
   sw_break();
#endif

   // All flows automatically started at 0 in my NetGraph class
   // While a path exists to the sink in the residual network 
   while (find_path(residual, sink)) {
      // Find residual capacity along the discovered path
      int rcapacity = INFINITY;
      int p, s;                  // p is predecessor vertex, s is successor
      for (s=sink, p=residual.pred[s]; p!=UNDEF; s=p, p=residual.pred[p]) {
         if (residual.weight(p, s) < rcapacity)
            rcapacity = residual.weight(p, s);
#ifdef SWAN
	 sw_setedgeattr(resid_swgraph, NODEID(p), NODEID(s), GEDGELINETH,
			THICKLINE); 	 // Mark path
#endif
      }
#ifdef SWAN
      memstm.seekp(0);
      memstm << "Path to sink found in residual network--residual capacity is "
	     << rcapacity << "." << ends;
      sw_print(memstr);
      sw_break();
#endif
      // Update edges on the discovered path
      for (s=sink, p=residual.pred[s]; p!=UNDEF ; s=p, p=residual.pred[p]) {
         // Add the residual capacity to the flow along path
         orig.add_flow(p, s, rcapacity);   // Adds to the correct edge
         // Update corresponding residual graph edges
         int newweight = residual.weight(p,s) - rcapacity;
         if (newweight > 0)
            residual.change_weight(p, s, newweight);
         else				   // Weight is <= 0--remove edge
            residual.remove_edge(p,s);
         // Add rcapacity to residual edge in other direction--create if nec.
	 if (residual.isEdge(s,p))
	    residual.change_weight(s,p, residual.weight(s,p) + rcapacity);
	 else
	    residual.add_edge(s,p,rcapacity);
      }      // end for each edge on discovered path
#ifdef SWAN
      sw_deletegraph(orig_swgraph);
      creategraph(orig, orig_swgraph);
      sw_displaygraph(orig_swgraph);
      sw_print("The flow has been augmented along the path in the original network.");
      sw_break();
      sw_deletegraph(resid_swgraph);	  // stay in same place
      creategraph(residual, resid_swgraph);
      sw_displaygraph(resid_swgraph);
      sw_print("The residual network has been updated.");
      sw_break();
#endif
   }      // end while still a path to sink on residual network

   int maxflow = 0;	// Now calculate the maximum flow
   for (Edge e = orig.first(0); orig.isEdge(e); e = orig.next(e))
      maxflow += NetEdge(e)->flow;
#ifdef SWAN
   sw_step();		   // Stop them here if they've chosed to run.
   memstm.seekp(0);
   memstm << "No path found in the residual network.  Maximum flow is "
	  << maxflow << "." << ends;
   sw_print(memstr);
   sw_break();
#else
   cout << "\nMaximum flow is " << maxflow << endl;
#endif
}

//////////////////////////////////////////////////
int GS_MAIN(GS_MAIN_ARGS_F)
//
// PURPOSE: Driver for program
// PARAMETERS: The standard ones
{
   char    temp_str[80] ;

   sw_init(GS_MAIN_ARGS_A);
   sw_print("What file would you like to use?") ;
   sw_getstr(temp_str) ;
   ifstream fs(temp_str);

   NetGraph maingraph(fs);    // Create graph from input
   if (!maingraph) { cout << "Unable to create graph.\n"; exit(-1);  }

   sw_step();

   // Find max flow with ford_fulkerson--use last vertex as sink
   ford_fulkerson(maingraph, maingraph.n() - 1);
   return 0;
}







   


