/*****************************************************************************
MODULE: topsort.cpp
PROJECT: topsort
AUTHOR: Jeff Nielsen
DATE: March 1996

DESCRIPTION: Performs a topological sort on a DAG.  Annotated with SAIL
	     function calls in order to view the topological sort using
	     SWAN.  Compile with the macro SWAN defined for the visualization.
USAGE: topsort [graph_file]
       If graph_file is not specified, a directed graph must be entered
       from stdin.  The format for entering a directed graph is as follows:
       <number_of_vertices>
       Directed
       <vertex_a> <vertex_b>
       <vertex_c> <vertex_d>
       ...

       A directed edge is inserted from vertex_a to vertex_b, etc.

*****************************************************************************/
#include <iostream.h>
#include <fstream.h>
#include <strstrea.h>
#include "stdjeff.h"
#include "grlist.h"	// Adjacency list graph implementation
#include "llist.ht"	// Linked list template

#define SWAN
#define VISITED 1
#define UNVISITED 0

list<int> vertexlist;	   // Global linked list for storing vertices

// Swan functions ---------------------------------------------------------
#ifdef SWAN
#include "sail.h"
// Names for two graphs in SWAN window
GRAPHID maingraph = GRAPHID(100);
GRAPHID vertlgraph = GRAPHID(101);
char memstr[80];
ostrstream memstm(memstr, 80);	    // Global in-memory stream for formatting

//////////////////////////////////////////////////
inline void make_swangraph(GRAPHID gid, NODETYPE nt, BOOLEAN dir_undir,
			   LAYOUTCOMPTYPE lctype)
// PURPOSE: Macro for making SWAN graphs
// PARAMETERS:
//    gid: The graphid you want for this graph
//    nt: The type of node you want the graph to have
//    dir_undir: DIRECTED or UNDIRECTED graph
//    lctype: The default layout component you want this graph to have.
{
   gid = sw_newgraph(gid, nt, dir_undir);
   sw_newlc(gid, LC_ID(1), lctype);
   sw_setcurlc(gid, LC_ID(1));
}   

//////////////////////////////////////////////////
void drawgraphs(Graph& g)
// PURPOSE: Draws the two SWAN graphs in the SWAN window
// PARAMETER;
//    g: The actual directed graph used in the program
{
   int v;
   // Draw maingraph
   for (v = 0; v < g.n(); v++) {
      memstm.seekp(0);
      memstm << v << ends;
      sw_insertnode(maingraph, NODEID(v), memstr);
   }
   for (v = 0; v < g.n(); v++)
      for (Edge w = g.first(v); g.isEdge(w); w = g.next(w) )
	 sw_insertedge(maingraph, NODEID(w->v1), NODEID(w->v2), NULL);

   // Draw vertlgraph
   for (v = g.n(); v > 0; v--) 	    // Go backwards
      sw_insertnode(vertlgraph, NODEID(v), " ");
   for (v = g.n()-1; v > 0; v--)
      sw_insertedge(vertlgraph, NODEID(v+1), NODEID(v), NULL);
   
   sw_displayallgraphs();
}

//////////////////////////////////////////////////
void addto_listgraph(int v)
// PURPOSE: Labels new head of linked list with vertex number
// PARAMETER:
//    v: The vertex number
{
   static int nodecount;   // Which node you're on in the graphical list
   memstm.seekp(0);
   memstm << v << ends;
   sw_setnodeattr(vertlgraph, NODEID(++nodecount), GNODELABEL, memstr);
}

#endif	       // #ifdef SWAN


// Topological Sort functions -------------------------------------------

//////////////////////////////////////////////////
void tophelp(Graph& G, int v)
// PURPOSE:  Topsort helper function--does the actual work.  This function
//           (without the SAIL calls, of course) comes mostly from
//           Dr. Shaffer's book.
// PARAMETERS:
//    G: The graph being traversed
//    v: The current vertex
{
  G.mark[v] = VISITED;
#ifdef SWAN
  // Color "gray" for PreVisit
  sw_setnodeattr(maingraph, NODEID(v), GNODEFILLED, TRUE);
  sw_setnodeattr(maingraph, NODEID(v), GNODECOLOR, LGRAY);
  memstm.seekp(0);
  memstm << "Vertex " << v << " discovered." << ends;
  sw_print(memstr);
  sw_break();
#endif

  for (Edge w = G.first(v); G.isEdge(w); w = G.next(w))
     if (G.mark[w->v2] == UNVISITED) {
#ifdef SWAN
	sw_setedgeattr(maingraph, NODEID(v), NODEID(w->v2),
		       GEDGELINETH, THICKLINE);	   // Highlight search tree
#endif
	tophelp(G, w->v2);
     }
  vertexlist.setFirst();
  vertexlist.insert(v);	   // Add to head of real linked list
  
#ifdef SWAN
  // Now color "black" for PostVisit
  sw_setnodeattr(maingraph, NODEID(v), GNODECOLOR, MGRAY);
  addto_listgraph(v);	   // Add to head of graphical linked list
  memstm.seekp(0);
  memstm << "Finished with vertex " << v << "--add to head of linked list."
	 << ends;
  sw_print(memstr);
  sw_break();
#endif
}

//////////////////////////////////////////////////
void topsort(Graph& G)
// PURPOSE: Recursive topological sort function--from Dr. Shaffer's book.
// PARAMETER:
//    G: The graph to be sorted
{
   int i;
   for (i=0; i<G.n(); i++)	 // Initialize Mark array
      G.mark[i] = UNVISITED;
   for (i=0; i<G.n(); i++)	 // Process all vertices
      if (G.mark[i] == UNVISITED)
	 tophelp(G, i);          // Call recursive helper function
}

//////////////////////////////////////////////////
int GS_MAIN(GS_MAIN_ARGS_F)
// PURPOSE: Driver for program
// PARAMETERS: The standard ones
{
   char    temp_str[80] ;
	

#ifdef SWAN
   sw_init(GS_MAIN_ARGS_A);
   sw_step();

   sw_print("What file would you like to use?") ;
   sw_getstr(temp_str) ;
   ifstream fs(temp_str);
   if (!fs) { sw_print("Couldn't open file "); exit(-1); }
   Graph thegraph(fs);    // Create graph from input
   if (!thegraph) { cout << "Unable to create graph.\n"; exit(-1);  }

   sw_disablebuttons(RUN);
   sw_step();
   make_swangraph(maingraph, CIRCLE, DIRECTED, CIRCLENET);
   sw_setgraphattr(maingraph, GEDGELABEL, FALSE);

   make_swangraph(vertlgraph, BOX, DIRECTED, LISTACROSS);
   sw_setgraphattr(vertlgraph, GEDGELABEL, FALSE);
   sw_setgraphattr(vertlgraph, GLAYOUTMODE, MANULAYOUT);
   sw_setgraphpos(vertlgraph, 10, 380);

   drawgraphs(thegraph);
   sw_print("Press STEP to view the topological sort.");
   sw_break();
#endif

   topsort(thegraph);
   
#ifdef SWAN
   sw_print("The linked list now shows the topological sort.");
   sw_break();
#else
   // Print out vertexlist
   for (vertexlist.setFirst(); vertexlist.isInList(); vertexlist.next())
      cout << vertexlist.currValue() << " ";
   cout << endl;
#endif
   return 0;
}








